GtkWidget* create_sheets_main_dialog           (void);
GtkWidget* create_sheets_new_dialog            (void);
GtkWidget* create_sheets_edit_dialog           (void);
GtkWidget* create_sheets_remove_dialog         (void);
