EXPORTS
; exported functions

 apply_textattr_properties
 apply_textstr_properties
 arrow_bbox
 dia_arrow_draw
 arrow_type_from_name
 arrow_get_name_from_type
 get_arrow_names
 dia_arrow_load
 dia_arrow_save

 attribute_first_data
 attribute_num_data
 attributes_default_fgbg
 attributes_get_background
 attributes_get_default_end_arrow
 attributes_get_default_font
 attributes_get_default_line_style
 attributes_get_default_linewidth
 attributes_get_default_start_arrow
 attributes_get_foreground
 attributes_set_background
 attributes_set_default_end_arrow
 attributes_set_default_font
 attributes_set_default_line_style
 attributes_set_default_linewidth
 attributes_set_default_start_arrow
 attributes_set_foreground
 attributes_swap_fgbg
 new_attribute

 bezier_draw_control_lines
 dia_renderer_bezier_stroke
 dia_renderer_bezier_fill
 beziercommon_set_points
 beziercommon_closest_segment

 bezierconn_add_segment
 bezierconn_closest_handle
 bezierconn_closest_major_handle
 bezierconn_copy
 bezierconn_destroy
 bezierconn_distance_from
 bezierconn_init
 bezierconn_load
 bezierconn_move
 bezierconn_move_handle
 bezierconn_remove_segment
 bezierconn_save
 bezierconn_set_corner_type
 bezierconn_update_boundingbox
 bezierconn_update_data

 beziershape_add_segment
 beziershape_closest_handle
 beziershape_closest_major_handle
 beziershape_copy
 beziershape_destroy
 beziershape_distance_from
 beziershape_init
 beziershape_load
 beziershape_move
 beziershape_move_handle
 beziershape_remove_segment
 beziershape_save
 beziershape_set_corner_type
 beziershape_update_boundingbox
 beziershape_update_data

 bernstein_develop
 bezier_eval
 bezier_eval_tangent

 calculate_arrow_point

 dia_object_change_list_get_type
 dia_object_change_list_new
 dia_object_change_list_add

 color_convert
 color_equals
 color_new_rgb
 color_new_rgba

 composite_add_attribute
 composite_find_attribute

 connection_adjust_for_autogap
 connection_copy
 connection_destroy
 connection_init
 connection_load
 connection_move_handle
 connection_save
 connection_update_boundingbox
 connection_update_handles

 connpoint_is_autogap
 connpoint_update

 connpointline_add_points
 connpointline_adjust_count
 connpointline_can_add_point
 connpointline_can_remove_point
 connpointline_copy
 connpointline_create
 connpointline_destroy
 connpointline_load
 connpointline_putonaline
 connpointline_remove_points
 connpointline_save
 connpointline_update

 create_standard_arc
 create_standard_bezierline
 create_standard_beziergon
 create_standard_box
 create_standard_ellipse
 create_standard_group
 create_standard_image
 create_standard_path
 create_standard_path_from_list
 create_standard_path_from_object
 create_standard_path_from_text
 create_standard_polygon
 create_standard_polyline
 create_standard_text
 create_standard_zigzagline

 data_add_boolean
 data_add_color
 data_add_composite
 data_add_enum
 data_add_dict
 data_add_filename
 data_add_font
 data_add_int
 data_add_layer
 data_add_layer_at
 data_add_pixbuf
 data_add_point
 data_add_bezpoint
 data_add_real
 data_add_string
 data_add_text
 data_boolean
 data_color
 data_remove_layer
 data_dict
 data_emit
 data_enum
 data_filename
 data_font

 diagram_data_clone
 diagram_data_clone_selected
 data_foreach_object
 data_get_sorted_selected
 data_get_sorted_selected_remove
 data_highlight_add
 data_highlight_remove
 data_object_get_highlight
 data_int
 data_layer_count
 data_layer_get_index
 data_layer_get_nth
 data_lower_layer
 data_next
 data_pixbuf
 data_point
 data_bezpoint
 data_raise_layer
 data_real
 data_remove_all_selected
 data_render
 data_render_paginated
 data_select
 data_set_active_layer
 dia_diagram_data_get_active_layer
 data_string
 data_text
 data_unselect
 data_update_extents

 destroy_object_list

 dia_asin
 dia_acos

 dia_arrow_chooser_new
 dia_arrow_chooser_set_arrow
 dia_arrow_chooser_get_arrow_type
 dia_arrow_chooser_get_type

 dia_arrow_selector_get_arrow
 dia_arrow_selector_get_type
 dia_arrow_selector_new
 dia_arrow_selector_set_arrow

 dia_arrow_preview_new
 dia_arrow_preview_get_type
 dia_arrow_preview_set_arrow
 dia_arrow_preview_get_arrow

 dia_arrow_cell_renderer_new

 dia_assert_true
 dia_colour_selector_get_colour
 dia_colour_selector_get_type
 dia_colour_selector_new
 dia_colour_selector_set_colour
 dia_colour_selector_set_use_alpha

 dia_colour_cell_renderer_new

 dia_config_filename

 dia_context_add_message
 dia_context_add_message_with_errno
 dia_context_get_filename
 dia_context_new
 dia_context_release
 dia_context_reset
 dia_context_set_filename

 dia_font_ascent
 dia_font_build_layout
 dia_font_descent
 dia_font_get_context
 dia_font_get_description
 dia_font_get_family
 dia_font_get_height
 dia_font_get_legacy_name
 dia_font_get_psfontname
 dia_font_get_size
 dia_font_get_slant_string
 dia_font_get_style
 dia_font_get_weight_string
 dia_font_new
 dia_font_new_from_legacy_name
 dia_font_new_from_style
 dia_font_set_any_family
 dia_font_set_height
 dia_font_set_slant
 dia_font_set_slant_from_string
 dia_font_set_weight
 dia_font_set_weight_from_string
 dia_font_copy
 dia_font_string_width

 dia_font_selector_get_font
 dia_font_selector_get_type
 dia_font_selector_new
 dia_font_selector_set_font

 dia_guide_new
 dia_guide_copy
 dia_guide_free
 dia_guide_get_type

 dia_get_data_directory
 dia_get_lib_directory
 dia_get_locale_directory
 dia_relativize_filename
 dia_absolutize_filename

 dia_image_add_ref
 dia_image_filename
 dia_image_get_broken
 dia_image_get_mime_type
 dia_image_height
 dia_image_load
 dia_image_save
 dia_image_mask_data
 dia_image_unref
 dia_image_rgb_data
 dia_image_rgba_data
 dia_image_rowstride
 dia_image_width
 dia_image_pixbuf
 dia_image_new_from_pixbuf

 dia_import_renderer_get_type
 dia_import_renderer_get_objects

 pixbuf_encode_base64
 pixbuf_decode_base64

 dia_path_renderer_get_type
 path_build_ellipse

 dia_pattern_new
 dia_pattern_add_color
 dia_pattern_foreach
 dia_pattern_get_fallback_color
 dia_pattern_get_points
 dia_pattern_get_radius
 dia_pattern_get_settings
 dia_pattern_set_pattern
 dia_pattern_set_point
 dia_pattern_set_radius
 data_add_pattern
 data_pattern
 dia_object_set_pattern

 dia_line_style_selector_get_linestyle
 dia_line_style_selector_get_type
 dia_line_style_selector_new
 dia_line_style_selector_set_linestyle

 dia_line_chooser_new
 dia_line_chooser_get_type
 dia_line_chooser_set_line_style

 dia_line_preview_new
 dia_line_preview_get_type
 dia_line_preview_set_style

 dia_line_cell_renderer_new

 dia_list_plugins

 dia_message_filename

 dia_object_get_parent_layer
 dia_object_is_selected

 dia_object_default_create
 dia_object_default_get
 dia_object_defaults_load
 dia_object_defaults_save
 dia_object_is_selectable
 dia_object_sanity_check

 dia_object_set_meta
 dia_object_get_meta
 dia_object_set_pixbuf
 dia_object_set_string
 dia_object_get_num_connections


 dia_object_draw
 dia_object_distance_from
 dia_object_select
 dia_object_clone
 dia_object_move
 dia_object_move_handle
 dia_object_get_editor
 dia_object_apply_editor
 dia_object_get_menu
 dia_object_describe_properties
 dia_object_get_properties
 dia_object_set_properties
 dia_object_apply_properties
 dia_object_edit_text
 dia_object_transform

 dia_change_get_type
 dia_change_new
 dia_change_ref
 dia_change_unref
 dia_change_apply
 dia_change_revert

 dia_object_change_get_type
 dia_object_change_new
 dia_object_change_ref
 dia_object_change_unref
 dia_object_change_apply
 dia_object_change_revert

 dia_plugin_can_unload
 dia_plugin_check_version
 dia_plugin_get_description
 dia_plugin_get_filename
 dia_plugin_get_inhibit_load
 dia_plugin_get_name
 dia_plugin_get_symbol
 dia_plugin_info_init
 dia_plugin_is_loaded
 dia_plugin_load
 dia_plugin_set_inhibit_load
 dia_plugin_unload
 dia_pluginrc_write
 dia_register_builtin_plugin
 dia_register_plugin
 dia_register_plugins
 dia_register_plugins_in_dir

 dia_font_get_type
 dia_object_draw

 dia_renderer_fill_arc
 dia_renderer_get_type
 dia_renderer_get_text_width
 dia_renderer_get_font
 dia_renderer_set_font
 dia_renderer_set_fillstyle
 dia_renderer_set_linewidth
 dia_renderer_set_linestyle
 dia_renderer_set_linecaps
 dia_renderer_set_linejoin
 dia_renderer_set_pattern
 dia_renderer_draw_arc
 dia_renderer_draw_arc_with_arrows
 dia_renderer_draw_object
 dia_renderer_draw_line
 dia_renderer_draw_line_with_arrows
 dia_renderer_draw_string
 dia_renderer_draw_polygon
 dia_renderer_draw_polyline
 dia_renderer_draw_polyline_with_arrows
 dia_renderer_draw_rounded_polyline
 dia_renderer_draw_rounded_polyline_with_arrows
 dia_renderer_draw_rounded_rect
 dia_renderer_draw_rotated_text
 dia_renderer_draw_rotated_image
 dia_renderer_draw_ellipse
 dia_renderer_draw_image
 dia_renderer_draw_bezier
 dia_renderer_draw_bezier_with_arrows
 dia_renderer_draw_beziergon
 dia_renderer_draw_rect
 dia_renderer_begin_render
 dia_renderer_end_render
 dia_renderer_is_capable_of

 dia_interactive_renderer_get_type
 dia_interactive_renderer_draw_pixel_line
 dia_interactive_renderer_get_width_pixels
 dia_interactive_renderer_get_height_pixels
 dia_interactive_renderer_draw_object_highlighted
 dia_interactive_renderer_fill_pixel_rect
 dia_interactive_renderer_set_size
 dia_interactive_renderer_draw_pixel_rect
 dia_interactive_renderer_clip_region_clear
 dia_interactive_renderer_clip_region_add_rect

 dia_size_selector_get_type
 dia_size_selector_new
 dia_size_selector_set_locked
 dia_size_selector_set_size
 dia_size_selector_get_size

 dia_svg_style_init
 dia_svg_style_copy
 dia_svg_parse_color
 dia_svg_parse_style
 dia_svg_parse_style_string
 dia_svg_parse_path
 dia_svg_parse_transform
 dia_svg_from_matrix
 dia_matrix_is_invertible

 dia_svg_renderer_get_type

 dia_transform_new
 dia_transform_length
 dia_transform_coords
 dia_untransform_length

 dia_transform_renderer_new

 dia_toggle_button_new_with_icon_names

 dia_sort_sheets
 get_sheets_list
 load_all_sheets
 register_sheet
 sheet_append_sheet_obj

 diagram_data_get_type

 dia_unit_spinner_get_type
 dia_unit_spinner_get_value
 dia_unit_spinner_new
 dia_unit_spinner_set_value
 dia_unit_spinner_set_upper

; dia_svg_parse_style

 distance_bez_line_point
 distance_bez_shape_point
 distance_ellipse_point
 distance_line_point
 distance_polygon_point
 distance_rectangle_point

 dynobj_list_add_object
 dynobj_list_foreach
 dynobj_list_get_dynobj_rate
 dynobj_list_remove_object


 element_copy
 element_destroy
 element_init
 element_load
 element_move_handle
 element_move_handle_aspect
 element_save
 element_update_boundingbox
 element_update_handles
 element_update_connections_rectangle
 element_update_connections_directions
 element_get_poly

 ellipse_bbox
 element_change_new

 filter_export_get_by_name
 filter_get_callbacks
 filter_get_export_filter_label
 filter_get_export_filters
 filter_set_favored_export
 filter_get_unique_export_names
 filter_get_import_filter_label
 filter_get_import_filters
 filter_guess_export_filter
 filter_guess_import_filter
 filter_import_get_by_name
 filter_register_callback
 filter_register_export
 filter_register_import
 filter_unregister_callback
 filter_unregister_export
 filter_unregister_import

 find_paper
 find_slope_directions

 focus_get_object
 focus_get_first_on_object
 focus_previous_on_diagram
 focus_next_on_diagram


 get_default_paper
 get_paper_bmargin
 get_paper_info
 get_paper_lmargin
 dia_paper_get_name
 get_paper_name_list
 get_paper_psheight
 get_paper_pswidth
 get_paper_rmargin
 get_paper_tmargin

 dia_line_caps_get_type
 dia_line_join_get_type
 dia_line_style_get_type
 dia_fill_style_get_type
 dia_alignment_get_type
 dia_text_fitting_get_type

 dia_unit_get_type
 dia_unit_get_name
 dia_unit_get_symbol
 dia_unit_get_factor
 dia_unit_get_digits

 group_create
 group_create_with_matrix
 group_get_transform
 group_destroy_shallow
 group_objects
 ;group_type

 get_active_focus
 give_focus

 intl_score_locale

 dia_layer_add_object
 dia_layer_add_object_at
 dia_layer_add_objects
 dia_layer_add_objects_first
 dia_layer_find_closest_connectionpoint
 dia_layer_find_closest_object
 dia_layer_find_closest_object_except
 dia_layer_find_objects_containing_rectangle
 dia_layer_find_objects_in_rectangle
 dia_layer_find_objects_intersecting_rectangle
 dia_layer_get_parent_diagram
 dia_layer_get_name
 dia_layer_object_count
 dia_layer_object_get_index
 dia_layer_object_get_nth
 dia_layer_remove_object
 dia_layer_remove_objects
 dia_layer_render
 dia_layer_replace_object_with_list
 dia_layer_set_object_list
 dia_layer_update_extents
 dia_layer_new
 dia_layer_new_from_layer
 dia_layer_is_visible
 dia_layer_set_visible
 dia_layer_is_connectable
 dia_layer_set_connectable
 dia_layer_get_type
 dia_layer_get_object_list
 dia_layer_get_extents

 line_bbox
 line_line_intersection

 message_error
 message_notice
 message_warning
 set_message_func

 dia_state_object_change_new

 object_add_connectionpoint
 object_add_handle
 object_apply_props
 object_apply_props_from_dialog
 object_connect
 object_copy
 object_copy_list
 object_copy_props
 object_copy_using_properties
 object_create_props_dialog
 object_describe_props
 object_destroy
 object_find_attribute
 object_flags_set
 object_get_prop_descriptions
 object_get_props
 object_get_props_from_offsets
 object_get_type
 object_get_displayname
 object_init
 object_list_move_delta
 object_list_create_props_dialog
 object_load
 object_load_props
 object_load_using_properties
 object_prop_by_name
 object_prop_by_name_type
 object_register_type
 object_registry_foreach
 object_remove_connectionpoint
 object_remove_connections_to
 object_remove_handle
 object_save
 object_save_props
 object_save_using_properties
 object_set_props_from_offsets
 object_substitute
 object_toggle_prop
 object_unconnect
 object_unconnect_all

 orthconn_add_segment
 orthconn_can_add_segment
 orthconn_can_delete_segment
 orthconn_copy
 orthconn_delete_segment
 orthconn_destroy
 orthconn_distance_from
 orthconn_get_middle_handle
 orthconn_init
 orthconn_load
 orthconn_move
 orthconn_move_handle
 orthconn_save
 orthconn_set_points
 orthconn_toggle_autorouting_callback
 orthconn_update_boundingbox
 orthconn_update_data
 orthconn_update_object_menu

 parent_handle_extents
 parent_handle_move_in_check
 parent_handle_move_out_check
 parent_list_affected
 parent_list_affected_hierarchy
 parent_move_child_delta

 pdtpp_true
 pdtpp_is_visible

 persistence_load
 persistence_save
 persistence_register_window
 persistence_register_window_create
 persistence_register_string_entry
 persistence_change_string_entry
 persistence_register_list
 persistent_list_clear
 persistent_list_get
 persistent_list_get_glist
 persistent_list_add
 persistent_list_set_max_length
 persistent_list_remove
 persistence_register_boolean
 persistence_register_color
 persistence_register_integer
 persistence_register_real
 persistence_register_string
 persistence_set_boolean
 persistence_set_color
 persistence_set_integer
 persistence_set_real
 persistence_set_string
 persistent_list_remove_all
 persistence_get_string
 persistence_get_integer
 persistence_get_real

 point_convex
 point_cross
 transform_length
 transform_point
 transform_bezpoint

 three_point_circle

 polybezier_bbox

 polyconn_add_point
 polyconn_closest_handle
 polyconn_closest_segment
 polyconn_copy
 polyconn_destroy
 polyconn_distance_from
 polyconn_init
 polyconn_load
 polyconn_move
 polyconn_move_handle
 polyconn_remove_point
 polyconn_save
 polyconn_set_points
 polyconn_update_boundingbox
 polyconn_update_data

 polyline_bbox

 polyshape_add_point
 polyshape_closest_handle
 polyshape_closest_segment
 polyshape_copy
 polyshape_destroy
 polyshape_distance_from
 polyshape_init
 polyshape_load
 polyshape_move
 polyshape_move_handle
 polyshape_remove_point
 polyshape_save
 polyshape_set_points
 polyshape_update_boundingbox
 polyshape_update_data

 make_new_prop
 prefs_get_length_unit
 prefs_set_length_unit
 prefs_get_fontsize_unit
 prefs_set_fontsize_unit
 prop_desc_find_real_handler
 prop_desc_insert_handler
 prop_desc_list_calculate_quarks
 prop_desc_list_find_prop
 prop_desc_list_free_handler_chain
 prop_desc_lists_intersection
 prop_list_add_list
 prop_list_add_filename
 prop_list_add_fill_colour
 prop_list_add_line_colour
 prop_list_add_line_style
 prop_list_add_line_width
 prop_list_add_matrix
 prop_list_add_point
 prop_list_add_real
 prop_list_add_string
 prop_list_add_show_background
 prop_list_add_font
 prop_list_add_fontsize
 prop_list_add_enum
 prop_list_add_text
 prop_list_add_text_colour
 prop_list_free
 prop_list_from_descs
 prop_list_from_single

 point_in_rectangle
 rectangle_add_point
 rectangle_bbox
 rectangle_in_rectangle
 rectangle_intersection
 rectangle_intersects
 rectangle_union
 rectangle_union

 remove_focus_on_diagram
 remove_focus_object

 reset_foci_on_diagram

 new_text
 new_text_default
 text_calc_boundingbox
 text_copy
 text_delete_all
 text_delete_key_handler
 text_destroy
 text_distance_from
 text_draw
 text_get_ascent
 text_get_attributes
 text_get_descent
 text_get_height
 text_get_line
 text_get_line_strlen
 text_get_line_width
 text_get_max_width
 text_get_string_copy
 text_grab_focus
 text_is_empty
 text_set_alignment
 text_set_attributes
 text_set_color
 text_set_cursor
 text_set_cursor_at_end
 text_set_font
 text_set_height
 text_set_position
 text_set_string

 text_line_new
 text_line_destroy
 text_line_copy
 text_line_set_string
 text_line_set_height
 text_line_set_font
 text_line_get_string
 text_line_get_font
 text_line_get_height
 text_line_calc_boundingbox_size
 text_line_get_width
 text_line_get_ascent
 text_line_get_descent
 text_line_get_alignment_adjustment

 dia_check_version
 dia_version_string

 diaXmlParseFile
 xmlDiaSaveFile
 xmlDoParseFile

 prop_get_data_from_widgets
 prop_dialog_from_widget
 find_prop_by_name
 dia_object_get_bounding_box
 dia_object_get_enclosing_box

 calculate_object_edge

; exported for libart renderer only
 dia_transform_coords_double
 text_line_adjust_layout_line

 libdia_init
 dia_log_message

 dia_marshal_VOID__OBJECT_UINT_POINTER

 dia_matrix_is_identity
 dia_matrix_get_angle_and_scales
 dia_matrix_set_angle_and_scales
 dia_matrix_set_rotate_around
 dia_matrix_multiply

 dia_option_menu_get_type
 dia_option_menu_add_item
 dia_option_menu_get_active
 dia_option_menu_new
 dia_option_menu_set_active

 pixbuf_from_resource
 dia_cairo_interactive_renderer_new
 dia_interactive_renderer_paint
 dia_object_type_get_icon
 dia_interactive_renderer_set_selection
 cairo_export_data
 cairo_print_callback
 dia_cairo_renderer_get_type
