/* internal */
WIDGET *_arrayprop_get_widget(ArrayProperty *prop, PropDialog *dialog);
void    _arrayprop_reset_widget(ArrayProperty *prop, WIDGET *widget);
void    _arrayprop_set_from_widget(ArrayProperty *prop, WIDGET *widget);
