EXPORTS
 ; all these exports are only required for pydia
 ddisplay_active
 ddisplay_add_update_all
 ddisplay_close
 ddisplay_flush
 ddisplay_resize_canvas
 ddisplay_scroll
 ddisplay_scroll_down
 ddisplay_scroll_left
 ddisplay_scroll_right
 ddisplay_scroll_up
 ddisplay_set_origo
 ddisplay_set_title
 ddisplay_show_all
 ddisplay_zoom
 new_display
 diagram_add_update
 diagram_add_update_all
 diagram_find_clicked_object
 diagram_find_closest_connectionpoint
 diagram_find_closest_handle
 diagram_flush
 diagram_get_sorted_selected
 diagram_get_sorted_selected_remove
 diagram_group_selected
 diagram_is_modified
 diagram_is_selected
 diagram_load
 dia_diagram_new
 dia_diagram_set_file
 dia_diagram_get_file
 diagram_get_type
 diagram_remove_all_selected
 diagram_save
 diagram_select
 diagram_ungroup_selected
 diagram_unselect_object
 diagram_update_extents
 diagram_update_connections_object
 diagram_update_connections_selection
 dia_open_diagrams

 ; just for the applications
 app_init
 app_is_interactive
 app_splash_done
 toolbox_show

